/*
 * The MIT License
 *
 * Copyright 2021 Sebastian Schauer.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
package sudoku.controller;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import static java.lang.Integer.parseInt;
import java.util.ArrayList;
import java.util.List;
import java.util.Scanner;
import javax.swing.JFileChooser;
import javax.swing.filechooser.FileNameExtensionFilter;
import sudoku.model.board.Sudoku;
import sudoku.model.games.*;
import sudoku.view.Window;

/**
 * The controller of the application.
 * 
 * @author Sebastian Schauer
 */
public class GameController {

    Window window;
    
    /**
     * The constructor
     * @param size The size of the empty sudoku to be loaded at startup.
     */
    public GameController(int size) {
        // Create the window
        window = new Window(this);
        
        // Create an empty sudoku and show it
        SudokuTransport sudoku = new SudokuTransport(size);
        window.setSudoku(sudoku);
    }
    
    Thread solvingThread;
    /**
     * Starts solving a sudoku.
     * @param st The sudoku to be solved
     */
    public void solve(SudokuTransport st) {
        // Create a real sudoku from the transport object
        Sudoku s = Sudoku.getSudoku(st, this);
        // Solve it and return
        if (solvingThread != null) {
            solvingThread.interrupt();
            solvingThread = null;
        }
        Thread th = new Thread(() -> { s.solve(); });
        th.start();
    }
    
    /**
     * Invoked by the solving thread when done
     * @param s The solved sudoku
     * @param isSolved Whether a solution was found
     */
    public void gotSolution(Sudoku s, boolean isSolved) {
        if (isSolved)
            window.setSudoku(s.getSudokuTransport());
        else
            ;
    }
    
    /**
     * Loads a sudoku from file.
     * @param filename The filename of the file
     * @return The initialized sudoku, null in case of error
     */
    public SudokuTransport load(File lastFile) {
        // Create the JFileChooser dialog
        JFileChooser fc = new JFileChooser(lastFile);
        FileNameExtensionFilter restrict = new FileNameExtensionFilter("Sudoku files", "sud");
        fc.addChoosableFileFilter(restrict);
        // Show it
        int r = fc.showOpenDialog(fc);
        // Read file if no error
        if (r == JFileChooser.APPROVE_OPTION) {
            return read(fc.getSelectedFile());
        }
        else {
            return null;
        }
    }
    
    /**
     * Reads a file from disk and creates a sudoku transport object
     * @param file The file to be read
     * @return The sudoku transport object
     */
    private SudokuTransport read(File file) {
        // Initialisation
        SudokuTransport sudoku = null;
        
        try {
            // Create a scanner
            Scanner scanner = new Scanner(file);
            // Read file into text
            List<String> text = new ArrayList<>();
            while (scanner.hasNext()) {
                text.add(scanner.next());
            }
            // First line: Read size of the sudoku
            int size = parseInt(text.get(0));
            // Create empty sudoku transport object
            sudoku = new SudokuTransport(size);
            // Read sudoku type from second line
            switch (text.get(1)) {
                case "Color":
                    sudoku.type = GameType.Color;
                    break;
                case "Hyper":
                    sudoku.type = GameType.Hyper;
                    break;
                case "Regular":
                    sudoku.type = GameType.Regular;
                    break;
                case "X":
                    sudoku.type = GameType.X;
                    break;
                default:
                    System.err.println("Unknown sudoku type!");
                    return null;
            }
            // Read sudoku values
            for (int i = 0; i < size * size; i++) {
                if (!"0".equals(text.get(i + 2))) {
                    sudoku.values[i / size][i % size] = parseInt(text.get(i + 2));
                }
            }
        } catch (Exception ex) {
            // In case of an error return null
            System.out.println(ex);
            return null;
        }
        // Return sudoku
        return sudoku;
    }
    
    /**
     * Saves the sudoku to file. Does not ask for overwriting.
     * @param filename The filename of the file
     * @return True in case of success, false otherwise
     */
    public File save(SudokuTransport st, File file) {
        // Prepare the file chooser dialog
        JFileChooser fc = new JFileChooser(file);
        FileNameExtensionFilter restrict = new FileNameExtensionFilter("Sudoku files", "sud");
        fc.addChoosableFileFilter(restrict);
        // Show it
        int r = fc.showSaveDialog(null);
        // Process in case of correct input
        if (r == JFileChooser.APPROVE_OPTION)
            // Write to file
            if (write(st, fc.getSelectedFile())) {
                // Return the file (for saving as last file)
                return fc.getSelectedFile();
            }
        // Error
        return null;
    }
    
    /**
     * Writes a sudoku transport object to disk. Does not ask for overwriting.
     * @param st The sudoku object
     * @param file The file to write into
     * @return Whether the writing was successful
     */
    public boolean write(SudokuTransport st, File file) {
        // Initialisation
        boolean success = true;
        try {
            // Build the text for writing: size, type
            String txt = st.size + "\n" + st.type.name() + "\n";
            // Write the values
            for (int x = 0; x < st.size; x++) {
                for (int y = 0; y < st.size; y++) {
                    txt += st.values[x][y] + " ";
                }
                txt += "\n";
            }
            // Write to file
            try (BufferedWriter writer = new BufferedWriter(new FileWriter(file))) {
                writer.write(txt);
            }
            catch (Exception ex) {
                success = false;
                System.out.println(ex);
            }
        } catch (Exception ex) {
            success = false;
            System.out.println(ex);
        }
        return success;
    }
    
    /**
     * Creates an empty sudoku transport object of given size and type.
     * @param size The desired size of the sudoku transport object
     * @param type The desired type of the sudoku transport object
     * @return The empty sudoku transport object
     */
    public SudokuTransport newSudoku(int size, GameType type) {
        SudokuTransport st = new SudokuTransport(size);
        st.type = type;
        return st;
    }
}
