/*
 * The MIT License
 *
 * Copyright 2021 Sebastian Schauer.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
package sudoku.view;

import java.awt.BasicStroke;
import java.awt.Font;
import java.awt.Insets;
import java.awt.event.FocusEvent;
import java.awt.event.FocusListener;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import javax.swing.BorderFactory;
import javax.swing.JButton;
import javax.swing.border.Border;

/**
 * A sudoku cell
 * 
 * @author Sebastian Schauer
 */
public class SudokuCell extends JButton {
    
    /**
     * The value of the cell (0 if empty, 10 in case of 'A')
     */
    int value;
    /**
     * The number of allowed numbers
     */
    int size;
    /**
     * The SudokuView the cell belongs to
     */
    SudokuView myView;
    /**
     * The x coordinate of the cell
     */
    int x;
    /**
     * The y coordinate of the cell
     */
    int y;
    /**
     * The border style of an unmarked cell
     */
    Border defaultBorder;
    
    public SudokuCell(int size, SudokuView view, int x, int y) {
        myView = view;
        this.size = size;
        this.setFont(new Font(this.getFont().getFontName(), this.getFont().getStyle(), 200 / size));
        this.x = x;
        this.y = y;
        this.defaultBorder = this.getBorder();
        this.setMargin(new Insets(0,0,0,0));
        this.addKeyListener(new KeyListener() {
            @Override
            public void keyTyped(KeyEvent arg0) { }
            @Override
            public void keyPressed(KeyEvent arg0) {
                char c = arg0.getKeyChar();
                if (c >= '0' && c <= '9') {
                    if (setValue(c - '0'))
                        myView.setChanged(true);
                    return;
                }
                if (c >= 'a' && c <= 'z') { 
                    if (setValue(c - 'a' + 10))
                        myView.setChanged(true);
                    return;
                }
                if (c >= 'A' && c <= 'Z') { 
                    if (setValue(c - 'A' + 10))
                        myView.setChanged(true);
                    return;
                }
                int keyCode = arg0.getKeyCode();
                switch(keyCode) {
                    case KeyEvent.VK_RIGHT: myView.goRight(x, y); break;
                    case KeyEvent.VK_LEFT: myView.goLeft(x, y); break;
                    case KeyEvent.VK_UP: myView.goUp(x, y); break;
                    case KeyEvent.VK_DOWN: myView.goDown(x, y); break;
                }
            }
            @Override
            public void keyReleased(KeyEvent arg0) {}
        });
        this.addFocusListener(new FocusListener() {
            @Override
            public void focusGained(FocusEvent arg0) {
                setBorder(BorderFactory.createStrokeBorder(new BasicStroke(3)));
            }
            @Override
            public void focusLost(FocusEvent arg0) {
                setBorder(defaultBorder);
            }
        });
    }
    
    /**
     * Sets the value of the cell. Displays 0 as '', 9 as '9' and 10 as 'A'
     * @param val The value to set
     * @return Whether it was successful or not
     */
    public boolean setValue(int val) {
        if (val > size)
            return false;
        value = val;
        if (value == 0)
            setText("");
        else
            if (val > 9)
                setText("" + (char)('A' + val - 10));
            else
                setText("" + val);
        return true;
    }
    
    /**
     * The value of the cell.
     * @return The value. 0 if empty, 10 in case of 'A'
     */
    public int getValue() {
        return value;
    }
}
